module crypted.digests.mdcoreimpl;

import string; /* memcpy / memset */

import crypted.digests.md;

class PaddedMessageDigest : MessageDigest
{
protected:
	ulong bitLength;
	int buffered;
	byte[] buffer;

// engine methods;

	abstract void   coreInit();
	abstract void   coreUpdate( byte[] buf );
	abstract byte[] coreDigest();

	abstract int  stateBufferByteLength();
	abstract byte getPaddingByte();
	abstract bit  bigEndianBitLength();
	abstract int  lengthOfBitLength();
	
public:
	void /* final */
	init()
	{
		bitLength = 0;
		buffered  = 0;
		coreInit();
	}

	void /* final */
	update( byte * buf, int length )
	in
	{
		assert( buffer.length == stateBufferByteLength() );
	}
	out
	{
		assert( (cast(uint)buffered) < buffer.length );
	}
	body
	{
		int blklen, tocpy;
		if ( length == 0 ) { return; }
		assert( length > 0 );
		bitLength += (length * 8);
		blklen = buffer.length;
		if ( length >= (tocpy = (blklen - buffered)) )
		{
			memcpy( &(buffer[buffered]), buf, tocpy );
			buf += tocpy;
			length -= tocpy;
			buffered = 0;
			coreUpdate( buffer );
			while( length >= blklen )
			{
				coreUpdate( buf[0..blklen] );
				buf += blklen;
				length -= blklen;
			}
		}

		memcpy( &(buffer[buffered]), buf, length );
		buffered += length;
	}

	byte[] /* final */
	digest()
	in
	{
		assert( (cast(uint)buffered) < buffer.length );
		assert( buffer.length == stateBufferByteLength() );
	}
	out( result )
	{
		assert( (result.length * 8) == digestBitLength() );
	}
	body
	{
		int bitlenLength, minPos;
		bitlenLength = lengthOfBitLength();
		assert( (bitlenLength == 64) || (bitlenLength == 128) );
		buffer[buffered] = getPaddingByte();
		buffered++;
		minPos = (buffer.length - (bitlenLength>>>3));
		if ( buffered >= minPos )
		{
			if ( buffered < buffer.length )
			{
				memset( &(buffer[buffered]), 0, buffer.length - buffered );
			}
			else
			{
				assert( buffered == buffer.length );
			}
			coreUpdate( buffer );
			buffered = 0;
		}
		assert( buffered <= minPos );
		while( buffered < minPos )
		{
			buffer[buffered] = 0;
			buffered++;
		}
		// only 64 or 128 bit lengths are supported by this code
		if ( bitlenLength == 128 )
		{
			for( int i = 8; i > 0; i-- )
			{
				buffer[buffered] = 0;
				buffered++;
			}
		}
		if ( bigEndianBitLength() )
		{
			for( int i = 7; i >= 0; i-- )
			{
				buffer[buffered] = cast(byte)((bitLength >>> (i<<3))&0xFF);
				buffered++;
			}
		}
		else
		{
			for( int i = 0; i < 8; i++ )
			{
				buffer[buffered] = cast(byte)((bitLength >>> (i<<3))&0xFF);
				buffered++;
			}
		}
		assert( buffered == buffer.length );
		coreUpdate( buffer );
		return coreDigest();
	}

protected:
	this()
	{
		super();
		buffer = new byte[stateBufferByteLength()];
		init();
	}
}

class PaddedStandardMessageDigest : PaddedMessageDigest
{
protected:
	abstract bit useBigEndianDigest();
	abstract int[] getInternalContext();

	byte[] coreDigest()
	{
		byte[] buf = new byte[digestBitLength()>>>3];
		int [] ctx = getInternalContext();
		byte * ptr = &(buf[0]);
		assert( (ctx.length << 2) == buf.length );
		if ( useBigEndianDigest() )
		{
			for (int i = 0; i < ctx.length; i++)
			{
				*ptr++ = (byte)(ctx[i] >>> 24);
				*ptr++ = (byte)(ctx[i] >>> 16);
				*ptr++ = (byte)(ctx[i] >>> 8);
				*ptr++ = (byte)(ctx[i]);
			}
		}
		else
		{
			for (int i = 0; i < ctx.length; i++)
			{
				*ptr++ = (byte)(ctx[i]);
				*ptr++ = (byte)(ctx[i] >>> 8);
				*ptr++ = (byte)(ctx[i] >>> 16);
				*ptr++ = (byte)(ctx[i] >>> 24);
			}
		}
		return buf;
	}
}

